import * as dgram from "dgram";
import { WaveWriter, normalize_l16, normalize_l24 } from "./biw-utils.js";
import { FFT } from "./fft.js";
const RTP_HEADERSIZE = 12;
export class AudioStreamDistributor {
    cbs;
    constructor() {
        this.cbs = new Set();
    }
    add(cb) {
        this.cbs.add(cb);
    }
    remove(cb) {
        this.cbs.delete(cb);
    }
    process(data, seqnr, ts) {
        this.cbs.forEach((cb) => cb(data, seqnr, ts));
    }
    chain() {
        return (data, seqnr, ts) => this.process(data, seqnr, ts);
    }
}
export class AudioStreamReceiver {
    num_channels;
    format;
    m_socket = null;
    m_lastSeqNr;
    m_lastRtpTS;
    m_rtpTSErrors = 0;
    m_seqNrErrors = 0;
    on_data;
    on_error;
    constructor(num_channels, format) {
        this.num_channels = num_channels;
        this.format = format;
    }
    get rtpTSErrors() { return this.m_rtpTSErrors; }
    get seqNrErrors() { return this.m_seqNrErrors; }
    start(addrs) {
        if (this.m_socket)
            throw new Error("Already open");
        return new Promise((resolve, reject) => {
            const sentinel = setTimeout(() => reject(`Failed to open stream listener with local address ${addrs.src}, destination address ${addrs.dst}`), 10000);
            console.log(`Listening on address ${addrs.dst} port ${addrs.port}`);
            const socket = dgram.createSocket({ type: "udp4", reuseAddr: true });
            this.m_socket = socket;
            this.m_socket.bind(addrs.port, () => {
                try {
                    socket.addMembership(addrs.dst);
                    socket.addListener("error", (err) => {
                        if (this.on_error)
                            this.on_error(err);
                    });
                    socket.addListener("close", () => {
                        if (this.m_socket && this.on_error)
                            this.on_error(`Socket closed unexpectedly`);
                    });
                    switch (this.format) {
                        case "L16":
                            socket.addListener("message", (msg) => {
                                try {
                                    const rtp_hdr = this.handle_rtp_header(msg, 2);
                                    const channels = this.num_channels;
                                    function read_l16(sample) {
                                        const offset = RTP_HEADERSIZE + sample * channels * 2;
                                        return Array.from({ length: channels }, (_, idx) => msg.readInt16BE(offset + idx * 2));
                                    }
                                    if (this.on_data)
                                        this.on_data({
                                            hdr: {
                                                num_channels: channels,
                                                num_samples: rtp_hdr.num_samples,
                                                sample_format: "L16",
                                                sample_rate: 48000
                                            },
                                            read_raw: read_l16,
                                            read_normalized: (sample) => read_l16(sample).map(normalize_l16)
                                        }, rtp_hdr.seqnr, rtp_hdr.rtpts);
                                }
                                catch (e) {
                                    if (this.on_error)
                                        this.on_error(e);
                                }
                            });
                            break;
                        case "L24":
                            socket.addListener("message", (msg) => {
                                try {
                                    const rtp_hdr = this.handle_rtp_header(msg, 3);
                                    const channels = this.num_channels;
                                    function read_l24(sample) {
                                        const offset = RTP_HEADERSIZE + sample * channels * 3;
                                        return Array.from({ length: channels }, (_, idx) => msg.readIntBE(offset + idx * 3, 3));
                                    }
                                    if (this.on_data)
                                        this.on_data({
                                            hdr: {
                                                num_channels: channels,
                                                num_samples: rtp_hdr.num_samples,
                                                sample_format: "L24",
                                                sample_rate: 48000
                                            },
                                            read_raw: read_l24,
                                            read_normalized: (sample) => read_l24(sample).map(normalize_l24)
                                        }, rtp_hdr.seqnr, rtp_hdr.rtpts);
                                }
                                catch (e) {
                                    if (this.on_error)
                                        this.on_error(e);
                                }
                            });
                            break;
                        case "AM824":
                            socket.addListener("message", (msg) => {
                                try {
                                    const rtp_hdr = this.handle_rtp_header(msg, 4);
                                    const channels = this.num_channels;
                                    function read_am824(sample) {
                                        const offset = RTP_HEADERSIZE + sample * channels * 4 + 1;
                                        return Array.from({ length: channels }, (_, idx) => msg.readIntBE(offset + idx * 4, 3));
                                    }
                                    if (this.on_data)
                                        this.on_data({
                                            hdr: {
                                                num_channels: channels,
                                                num_samples: rtp_hdr.num_samples,
                                                sample_format: "L24",
                                                sample_rate: 48000
                                            },
                                            read_raw: read_am824,
                                            read_normalized: (sample) => read_am824(sample).map(normalize_l24)
                                        }, rtp_hdr.seqnr, rtp_hdr.rtpts);
                                }
                                catch (e) {
                                    if (this.on_error)
                                        this.on_error(e);
                                }
                            });
                            break;
                    }
                    clearTimeout(sentinel);
                    resolve();
                }
                catch (e) {
                    reject(e);
                }
            });
        });
    }
    handle_rtp_header(msg, bps) {
        const num_samples = Math.floor((msg.byteLength - RTP_HEADERSIZE) / (bps * this.num_channels));
        const seqnr = msg.readUInt16BE(2);
        const rtpts = msg.readUInt32BE(4);
        if (this.m_lastRtpTS !== undefined) {
            let delta = (rtpts - this.m_lastRtpTS) & 0xFFFFFFFF;
            if (delta != num_samples)
                this.m_rtpTSErrors++;
        }
        else
            console.log(`First packet Seq#=${seqnr.toString(16)} ts=0x${rtpts.toString(16)} num_samples=${num_samples} msg.byteLength=${msg.byteLength}`);
        if (this.m_lastSeqNr !== undefined) {
            let delta = (seqnr - this.m_lastSeqNr) & 0xFFFF;
            if (delta != 1)
                this.m_seqNrErrors++;
        }
        this.m_lastSeqNr = seqnr;
        this.m_lastRtpTS = rtpts;
        return { num_samples, seqnr, rtpts };
    }
    stop() {
        if (this.m_socket) {
            const socket = this.m_socket;
            this.m_socket = null;
            socket.close();
        }
    }
}
export class AudioStreamWriter {
    writer;
    hdr = null;
    constructor(path) {
        this.writer = new WaveWriter(path);
    }
    get num_samples() {
        return this.hdr?.num_samples ?? 0;
    }
    process(data) {
        this.writer.write_data(data);
        if (this.hdr) {
            this.hdr = {
                ...data.hdr,
                num_samples: data.hdr.num_samples + this.hdr.num_samples
            };
        }
        else
            this.hdr = data.hdr;
    }
    write_header() {
        if (this.hdr)
            this.writer.write_header(this.hdr);
    }
    chain() {
        return (data) => this.process(data);
    }
    close() {
        this.write_header();
        this.writer.close();
    }
}
export class AudioNormalizedRingBuffer {
    buf;
    pos;
    on_data;
    constructor(pars) {
        this.buf = Array.from({ length: pars.num_channels }, () => new Float32Array(pars.num_samples));
        this.pos = 0;
    }
    process(data) {
        const buf = this.buf;
        const len = buf[0].length;
        let pos = this.pos;
        for (let sample = 0; sample < data.hdr.num_samples; ++sample) {
            const samples = data.read_normalized(sample);
            samples.forEach((v, idx) => buf[idx][pos] = v);
            pos++;
            if (pos == len) {
                if (this.on_data) {
                    function read(sample) {
                        return buf.map((v) => v[sample]);
                    }
                    this.on_data(buf, {
                        hdr: {
                            num_samples: len,
                            num_channels: buf.length,
                            sample_format: "Float",
                            sample_rate: data.hdr.sample_rate
                        },
                        read_normalized: read,
                        read_raw: read
                    });
                }
                pos = 0;
            }
        }
        this.pos = pos;
    }
    reset() {
        this.pos = 0;
    }
    chain() {
        return (data, _seqnr, _rtpts) => this.process(data);
    }
    chain_reset_on_lost_packet(lost_cb) {
        let exp_seqnr = undefined;
        return (data, seqnr, _rtpts) => {
            if (exp_seqnr !== undefined && seqnr != exp_seqnr) {
                lost_cb?.(seqnr, exp_seqnr);
                this.reset();
            }
            exp_seqnr = (seqnr + 1) & 0xFFFF;
            this.process(data);
        };
    }
}
export class AudioFFT {
    fft;
    out;
    tmp;
    window;
    sample_rate;
    scale;
    on_data;
    constructor(pars) {
        this.fft = new FFT(pars.num_samples);
        this.out = Array.from({ length: pars.num_channels }, () => this.fft.createComplexArray());
        this.tmp = new Float64Array(pars.num_samples);
        this.window = new Float64Array(pars.num_samples);
        this.sample_rate = pars.sample_rate;
        const N = this.window.length;
        switch (pars.window) {
            case "Hann":
                for (let n = 0; n < N; ++n) {
                    const v = Math.sin(Math.PI * n / N);
                    this.window[n] = v * v;
                }
                break;
        }
        this.scale = 2.0 / this.window.reduce((a, b) => a + b);
    }
    process(data) {
        if (this.on_data) {
            const done = Array.from(this.out, () => false);
            const self = this;
            const N = this.tmp.length / 2;
            const freq_scale = this.sample_rate / (2 * N);
            function get_channel(ch) {
                const out = self.out[ch];
                const scale = self.scale;
                if (!done[ch]) {
                    const tmp = self.tmp;
                    const window = self.window;
                    data[ch].forEach((v, idx) => tmp[idx] = v * window[idx]);
                    self.fft.realTransform(out, tmp);
                    done[ch] = true;
                }
                function get_power(n) {
                    return Math.hypot(out[n * 2], out[n * 2 + 1]) * scale;
                }
                function find_peaks(threshold_dB) {
                    const threshold = threshold_dB !== undefined ? Math.pow(10, threshold_dB / 20.0) : undefined;
                    let ret = [];
                    let prev = get_power(1);
                    let start = null;
                    let accu = 0;
                    for (let n = 2; n < N; ++n) {
                        let cur = get_power(n);
                        if (prev < cur) {
                            start = n;
                            accu = cur;
                        }
                        else if (prev > cur) {
                            if (start !== null) {
                                const width = n - start;
                                const power = accu / width;
                                if (!threshold || power > threshold) {
                                    const mid = start + (width - 1) / 2;
                                    ret.push({
                                        n: mid,
                                        freq: mid * freq_scale,
                                        power,
                                        db: 20 * Math.log10(power)
                                    });
                                }
                            }
                            start = null;
                        }
                        else
                            accu += cur;
                        prev = cur;
                    }
                    return ret;
                }
                return {
                    get_power,
                    get_db: (n) => 20 * Math.log10(get_power(n)),
                    find_peaks
                };
            }
            this.on_data({
                length: N,
                channels: done.length,
                frequency: (n) => n * freq_scale,
                channel: get_channel
            });
        }
    }
    chain() {
        return (data) => this.process(data);
    }
}
